"""
Middleware to ensure best practices of DRF and other endpoints.
"""
import warnings

from django.utils.deprecation import MiddlewareMixin
from edx_django_utils import monitoring
from edx_django_utils.cache import DEFAULT_REQUEST_CACHE

from edx_rest_framework_extensions.auth.jwt.constants import USE_JWT_COOKIE_HEADER
from edx_rest_framework_extensions.auth.jwt.cookies import jwt_cookie_name


class RequestCustomAttributesMiddleware(MiddlewareMixin):
    """
    Adds various request related custom attributes.

    Possible custom attributes include:
        request_authenticated_user_set_in_middleware:
            Example values: 'process_request', 'process_view', 'process_response',
            or 'process_exception'. Attribute won't exist if user is not authenticated.
        request_auth_type_guess: Example values include: no-user, unauthenticated,
            jwt, bearer, other-token-type, jwt-cookie, or session-or-other
            Note: These are just guesses because if a token was expired, for example,
              the user could have been authenticated by some other means.
        request_client_name: The client name from edx-rest-api-client calls.
        request_referer
        request_user_agent: The user agent string from the request header.
        request_user_id: The user id of the request user.
        request_is_staff_or_superuser: `staff` or `superuser` depending on whether the
            user in the request is a django staff or superuser.

    This middleware is dependent on the RequestCacheMiddleware. You must
    include this middleware later.  For example::

        MIDDLEWARE = (
            'edx_django_utils.cache.middleware.RequestCacheMiddleware',
            'edx_rest_framework_extensions.middleware.RequestCustomAttributesMiddleware',
        )

    This middleware should also appear after any authentication middleware.

    """
    def process_request(self, request):
        """
        Caches if authenticated user was found.
        """
        self._cache_if_authenticated_user_found_in_middleware(request, 'process_request')

    def process_view(self, request, view_func, view_args, view_kwargs):  # pylint: disable=unused-argument
        """
        Caches if authenticated user was found.
        """
        self._cache_if_authenticated_user_found_in_middleware(request, 'process_view')

    def process_response(self, request, response):
        """
        Add custom attributes for various details of the request.
        """
        self._cache_if_authenticated_user_found_in_middleware(request, 'process_response')
        self._set_all_request_attributes(request)
        return response

    def process_exception(self, request, exception):  # pylint: disable=unused-argument
        """
        Django middleware handler to process an exception
        """
        self._cache_if_authenticated_user_found_in_middleware(request, 'process_exception')
        self._set_all_request_attributes(request)

    def _set_all_request_attributes(self, request):
        """
        Sets all the request custom attributes
        """
        self._set_request_auth_type_guess_attribute(request)
        self._set_request_user_agent_attributes(request)
        self._set_request_referer_attribute(request)
        self._set_request_user_id_attribute(request)
        self._set_request_authenticated_user_found_in_middleware_attribute()
        self._set_request_is_staff_or_superuser(request)

    def _set_request_is_staff_or_superuser(self, request):
        """
        Add `request_is_staff_or_superuser` custom attribute.

        Custom Attributes:
            request_is_staff_or_superuser
        """
        value = None
        if hasattr(request, 'user') and request.user:
            if request.user.is_superuser:
                value = 'superuser'
            elif request.user.is_staff:
                value = 'staff'

            if value:
                monitoring.set_custom_attribute('request_is_staff_or_superuser', value)

    def _set_request_user_id_attribute(self, request):
        """
        Add request_user_id custom attribute

        Custom Attributes:
             request_user_id
        """
        if hasattr(request, 'user') and hasattr(request.user, 'id') and request.user.id:
            monitoring.set_custom_attribute('request_user_id', request.user.id)

    def _set_request_referer_attribute(self, request):
        """
        Add custom attribute 'request_referer' for http referer.
        """
        if 'HTTP_REFERER' in request.META and request.META['HTTP_REFERER']:
            monitoring.set_custom_attribute('request_referer', request.META['HTTP_REFERER'])

    def _set_request_user_agent_attributes(self, request):
        """
        Add custom attributes for user agent for python.

        Custom Attributes:
             request_user_agent
             request_client_name: The client name from edx-rest-api-client calls.
        """
        if 'HTTP_USER_AGENT' in request.META and request.META['HTTP_USER_AGENT']:
            user_agent = request.META['HTTP_USER_AGENT']
            monitoring.set_custom_attribute('request_user_agent', user_agent)
            if user_agent:
                # Example agent string from edx-rest-api-client:
                #    python-requests/2.9.1 edx-rest-api-client/1.7.2 ecommerce
                #    See https://github.com/edx/edx-rest-api-client/commit/692903c30b157f7a4edabc2f53aae1742db3a019
                user_agent_parts = user_agent.split()
                if len(user_agent_parts) == 3 and user_agent_parts[1].startswith('edx-rest-api-client/'):
                    monitoring.set_custom_attribute('request_client_name', user_agent_parts[2])

    def _set_request_auth_type_guess_attribute(self, request):
        """
        Add custom attribute 'request_auth_type_guess' for the authentication type used.

        NOTE: This is a best guess at this point.  Possible values include:
            no-user
            unauthenticated
            jwt/bearer/other-token-type
            jwt-cookie
            session-or-other (catch all)

        """
        if not hasattr(request, 'user') or not request.user:
            auth_type = 'no-user'
        elif not request.user.is_authenticated:
            auth_type = 'unauthenticated'
        elif 'HTTP_AUTHORIZATION' in request.META and request.META['HTTP_AUTHORIZATION']:
            token_parts = request.META['HTTP_AUTHORIZATION'].split()
            # Example: "JWT eyJhbGciO..."
            if len(token_parts) == 2:
                auth_type = token_parts[0].lower()  # 'jwt' or 'bearer' (for example)
            else:
                auth_type = 'other-token-type'
        elif USE_JWT_COOKIE_HEADER in request.META and jwt_cookie_name() in request.COOKIES:
            auth_type = 'jwt-cookie'
        else:
            auth_type = 'session-or-other'
        monitoring.set_custom_attribute('request_auth_type_guess', auth_type)

    AUTHENTICATED_USER_FOUND_CACHE_KEY = 'edx-drf-extensions.authenticated_user_found_in_middleware'

    def _set_request_authenticated_user_found_in_middleware_attribute(self):
        """
        Add custom attribute 'request_authenticated_user_found_in_middleware' if authenticated user was found.
        """
        cached_response = DEFAULT_REQUEST_CACHE.get_cached_response(self.AUTHENTICATED_USER_FOUND_CACHE_KEY)
        if cached_response.is_found:
            monitoring.set_custom_attribute(
                'request_authenticated_user_found_in_middleware',
                cached_response.value
            )

    def _cache_if_authenticated_user_found_in_middleware(self, request, value):
        """
        Updates the cached process step in which the authenticated user was found, if it hasn't already been found.
        """
        cached_response = DEFAULT_REQUEST_CACHE.get_cached_response(self.AUTHENTICATED_USER_FOUND_CACHE_KEY)
        if cached_response.is_found:
            # since we are tracking the earliest point the authenticated user was found,
            # and the value was already set in earlier middleware step, do not set again.
            return

        if hasattr(request, 'user') and request.user and request.user.is_authenticated:
            DEFAULT_REQUEST_CACHE.set(self.AUTHENTICATED_USER_FOUND_CACHE_KEY, value)


class RequestMetricsMiddleware(RequestCustomAttributesMiddleware):
    """
    Deprecated class for handling middleware. Class has been renamed to RequestCustomAttributesMiddleware.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        msg = "Use 'RequestCustomAttributesMiddleware' in place of 'RequestMetricsMiddleware'."
        warnings.warn(msg, DeprecationWarning)



""" Tests for settings. """
import warnings
from unittest import mock

from django.conf import settings
from django.test import TestCase, override_settings

from edx_rest_framework_extensions.settings import get_jwt_issuers, get_setting


class SettingsTests(TestCase):
    """ Tests for settings retrieval. """

    @override_settings(EDX_DRF_EXTENSIONS={})
    def test_get_setting_with_missing_key(self):
        """ Verify the function raises KeyError if the setting is not defined. """
        self.assertRaises(KeyError, get_setting, 'not_defined')

    def test_get_setting(self):
        """ Verify the function returns the value of the specified setting from the EDX_DRF_EXTENSIONS dict. """

        _settings = {
            'some-setting': 'some-value',
            'another-one': False
        }

        with override_settings(EDX_DRF_EXTENSIONS=_settings):
            for key, value in _settings.items():
                self.assertEqual(get_setting(key), value)

    def test_get_current_jwt_issuers(self):
        """
        Verify the get_jwt_issuers operation returns the current issuer information when configured
        """
        self.assertEqual(get_jwt_issuers(), settings.JWT_AUTH['JWT_ISSUERS'])

    def test_get_deprecated_jwt_issuers(self):
        """
        Verify the get_jwt_issuers operation returns the deprecated issuer information when current
        issuers are not configured for the system.
        """
        _deprecated = [
            {
                'ISSUER': settings.JWT_AUTH['JWT_ISSUER'],
                'SECRET_KEY': settings.JWT_AUTH['JWT_SECRET_KEY'],
                'AUDIENCE': settings.JWT_AUTH['JWT_AUDIENCE'],
            }
        ]

        mock_call = 'edx_rest_framework_extensions.settings._get_current_jwt_issuers'
        with mock.patch(mock_call, mock.Mock(return_value=None)):
            with warnings.catch_warnings(record=True) as warning_list:
                warnings.simplefilter("default")
                self.assertEqual(get_jwt_issuers(), _deprecated)
                self.assertEqual(len(warning_list), 1)
                self.assertTrue(issubclass(warning_list[-1].category, DeprecationWarning))
                msg = "'JWT_ISSUERS' list not defined, checking for deprecated settings."
                self.assertIn(msg, str(warning_list[-1].message))



""" Utility functions for tests. """
from time import time

import jwt
from django.conf import settings


def generate_jwt(user, scopes=None, filters=None, is_restricted=None):
    """
    Generate a valid JWT for authenticated requests.
    """
    access_token = generate_latest_version_payload(
        user,
        scopes=scopes,
        filters=filters,
        is_restricted=is_restricted
    )
    return generate_jwt_token(access_token)


def generate_jwt_token(payload, signing_key=None):
    """
    Generate a valid JWT token for authenticated requests.
    """
    signing_key = signing_key or settings.JWT_AUTH['JWT_ISSUERS'][0]['SECRET_KEY']
    return jwt.encode(payload, signing_key)


def generate_latest_version_payload(user, scopes=None, filters=None, version=None,
                                    is_restricted=None):
    """
    Generate a valid JWT payload given a user and optionally scopes and filters.
    """
    payload = generate_unversioned_payload(user)
    payload.update({
        # fix this version and add newly introduced fields as the version updates.
        'version': '1.1.0',
        'filters': [],
        'is_restricted': False,
    })
    if scopes is not None:
        payload['scopes'] = scopes
    if version is not None:
        payload['version'] = version
    if filters is not None:
        payload['filters'] = filters
    if is_restricted is not None:
        payload['is_restricted'] = is_restricted
    return payload


def generate_unversioned_payload(user):
    """
    Generate an unversioned valid JWT payload given a user.
    """
    jwt_issuer_data = settings.JWT_AUTH['JWT_ISSUERS'][0]
    now = int(time())
    ttl = 600
    payload = {
        'iss': jwt_issuer_data['ISSUER'],
        'aud': jwt_issuer_data['AUDIENCE'],
        'username': user.username,
        'email': user.email,
        'iat': now,
        'exp': now + ttl,
        'scopes': [],
    }
    return payload



"""
Unit tests for middlewares.
"""
from unittest.mock import call, patch

import ddt
from django.contrib.auth.models import AnonymousUser
from django.test import RequestFactory, TestCase
from edx_django_utils.cache import RequestCache

from edx_rest_framework_extensions.auth.jwt.constants import USE_JWT_COOKIE_HEADER
from edx_rest_framework_extensions.auth.jwt.cookies import jwt_cookie_name
from edx_rest_framework_extensions.middleware import (
    RequestCustomAttributesMiddleware,
    RequestMetricsMiddleware,
)
from edx_rest_framework_extensions.tests.factories import UserFactory


@ddt.ddt
class TestRequestCustomAttributesMiddleware(TestCase):
    def setUp(self):
        super().setUp()
        RequestCache.clear_all_namespaces()
        self.request = RequestFactory().get('/')
        self.middleware = RequestCustomAttributesMiddleware()

    @patch('edx_django_utils.monitoring.set_custom_attribute')
    def test_request_auth_type_guess_anonymous_attribute(self, mock_set_custom_attribute):
        self.request.user = AnonymousUser()

        self.middleware.process_response(self.request, None)
        mock_set_custom_attribute.assert_called_once_with('request_auth_type_guess', 'unauthenticated')

    @patch('edx_django_utils.monitoring.set_custom_attribute')
    def test_request_no_headers(self, mock_set_custom_attribute):
        self.request.user = None
        self.middleware.process_response(self.request, None)
        mock_set_custom_attribute.assert_called_once_with('request_auth_type_guess', 'no-user')

    @patch('edx_django_utils.monitoring.set_custom_attribute')
    def test_request_blank_headers(self, mock_set_custom_attribute):
        self.request.META['HTTP_USER_AGENT'] = ''
        self.request.META['HTTP_REFERER'] = ''
        self.request.META['HTTP_AUTHORIZATION'] = ''

        self.middleware.process_response(self.request, None)
        mock_set_custom_attribute.assert_called_once_with('request_auth_type_guess', 'no-user')

    @patch('edx_django_utils.monitoring.set_custom_attribute')
    def test_request_referer_attribute(self, mock_set_custom_attribute):
        self.request.META['HTTP_REFERER'] = 'test-http-referer'

        self.middleware.process_response(self.request, None)
        expected_calls = [
            call('request_referer', 'test-http-referer'),
            call('request_auth_type_guess', 'no-user'),
        ]
        mock_set_custom_attribute.assert_has_calls(expected_calls, any_order=True)

    @patch('edx_django_utils.monitoring.set_custom_attribute')
    def test_request_rest_client_user_agent_attributes(self, mock_set_custom_attribute):
        self.request.META['HTTP_USER_AGENT'] = 'python-requests/2.9.1 edx-rest-api-client/1.7.2 test-client'

        self.middleware.process_response(self.request, None)
        expected_calls = [
            call('request_user_agent', 'python-requests/2.9.1 edx-rest-api-client/1.7.2 test-client'),
            call('request_client_name', 'test-client'),
            call('request_auth_type_guess', 'no-user'),
        ]
        mock_set_custom_attribute.assert_has_calls(expected_calls, any_order=True)

    @patch('edx_django_utils.monitoring.set_custom_attribute')
    def test_request_standard_user_agent_attributes(self, mock_set_custom_attribute):
        self.request.META['HTTP_USER_AGENT'] = 'test-user-agent'

        self.middleware.process_response(self.request, None)
        expected_calls = [
            call('request_user_agent', 'test-user-agent'),
            call('request_auth_type_guess', 'no-user'),
        ]
        mock_set_custom_attribute.assert_has_calls(expected_calls, any_order=True)

    @ddt.data(
        ('jwt abcdefg', 'jwt'),
        ('bearer abcdefg', 'bearer'),
        ('abcdefg', 'other-token-type'),
    )
    @ddt.unpack
    @patch('edx_django_utils.monitoring.set_custom_attribute')
    def test_request_auth_type_guess_token_attribute(self, token, expected_token_type, mock_set_custom_attribute):
        self.request.user = UserFactory()
        self.request.META['HTTP_AUTHORIZATION'] = token

        self.middleware.process_response(self.request, None)
        mock_set_custom_attribute.assert_any_call('request_auth_type_guess', expected_token_type)

    @patch('edx_django_utils.monitoring.set_custom_attribute')
    def test_request_auth_type_guess_jwt_cookie_attribute(self, mock_set_custom_attribute):
        self.request.user = UserFactory()
        self.request.META[USE_JWT_COOKIE_HEADER] = True
        self.request.COOKIES[jwt_cookie_name()] = 'reconstituted-jwt-cookie'

        self.middleware.process_response(self.request, None)
        mock_set_custom_attribute.assert_any_call('request_auth_type_guess', 'jwt-cookie')

    @patch('edx_django_utils.monitoring.set_custom_attribute')
    def test_request_auth_type_guess_session_attribute(self, mock_set_custom_attribute):
        self.request.user = UserFactory()

        self.middleware.process_response(self.request, None)
        mock_set_custom_attribute.assert_any_call('request_auth_type_guess', 'session-or-other')

    @patch('edx_django_utils.monitoring.set_custom_attribute')
    def test_request_user_id_attribute(self, mock_set_custom_attribute):
        self.request.user = UserFactory()

        self.middleware.process_response(self.request, None)
        mock_set_custom_attribute.assert_any_call('request_user_id', self.request.user.id)
        mock_set_custom_attribute.assert_any_call(
            'request_authenticated_user_found_in_middleware', 'process_response'
        )

    @patch('edx_django_utils.monitoring.set_custom_attribute')
    def test_request_user_id_attribute_with_exception(self, mock_set_custom_attribute):
        self.request.user = UserFactory()

        self.middleware.process_exception(self.request, None)
        mock_set_custom_attribute.assert_any_call('request_user_id', self.request.user.id)
        mock_set_custom_attribute.assert_any_call(
            'request_authenticated_user_found_in_middleware', 'process_exception'
        )

    @patch('edx_django_utils.monitoring.set_custom_attribute')
    def test_authenticated_user_found_in_process_request(self, mock_set_custom_attribute):
        self.request.user = UserFactory()
        self.middleware.process_request(self.request)
        self.middleware.process_response(self.request, None)

        mock_set_custom_attribute.assert_any_call(
            'request_authenticated_user_found_in_middleware', 'process_request'
        )

    @patch('edx_django_utils.monitoring.set_custom_attribute')
    def test_authenticated_user_found_in_process_view(self, mock_set_custom_attribute):
        self.request.user = UserFactory()
        self.middleware.process_view(self.request, None, None, None)
        self.middleware.process_response(self.request, None)

        mock_set_custom_attribute.assert_any_call(
            'request_authenticated_user_found_in_middleware', 'process_view'
        )

    @patch('edx_django_utils.monitoring.set_custom_attribute')
    def test_authenticated_user_found_is_properly_reset(self, mock_set_custom_attribute):
        # set user before process_request
        self.request.user = UserFactory()
        self.middleware.process_request(self.request)
        self.middleware.process_response(self.request, None)

        mock_set_custom_attribute.assert_any_call(
            'request_authenticated_user_found_in_middleware', 'process_request'
        )

        # set up new request and set user before process_response
        mock_set_custom_attribute.reset_mock()
        RequestCache.clear_all_namespaces()
        self.request = RequestFactory().get('/')
        self.request.user = UserFactory()
        self.middleware.process_response(self.request, None)

        mock_set_custom_attribute.assert_any_call(
            'request_authenticated_user_found_in_middleware', 'process_response'
        )

    @patch('edx_django_utils.monitoring.set_custom_attribute')
    def test_authenticated_standard_user(self, mock_set_custom_attribute):
        self.request.user = UserFactory()
        self.middleware.process_request(self.request)
        self.middleware.process_response(self.request, None)

        attributes_called_with = [c[0] for c in mock_set_custom_attribute.call_args_list]
        assert 'request_is_staff_or_superuser' not in attributes_called_with

    @patch('edx_django_utils.monitoring.set_custom_attribute')
    def test_authenticated_staff_user(self, mock_set_custom_attribute):
        self.request.user = UserFactory(is_staff=True)
        self.middleware.process_request(self.request)
        self.middleware.process_response(self.request, None)

        mock_set_custom_attribute.assert_any_call('request_is_staff_or_superuser', 'staff')

    @patch('edx_django_utils.monitoring.set_custom_attribute')
    def test_authenticated_superuser(self, mock_set_custom_attribute):
        self.request.user = UserFactory(is_superuser=True)
        self.middleware.process_request(self.request)
        self.middleware.process_response(self.request, None)

        mock_set_custom_attribute.assert_any_call('request_is_staff_or_superuser', 'superuser')


@ddt.ddt
class TestRequestMetricsMiddleware(TestCase):
    """
    Temporary smoke test of deprecated RequestMetricsMiddleware.

    The deprecated class is a subclass of RequestCustomAttributesMiddleware,
    which is more fully tested.
    """
    def setUp(self):
        super().setUp()
        RequestCache.clear_all_namespaces()
        self.request = RequestFactory().get('/')
        self.middleware = RequestMetricsMiddleware()

    @patch('edx_django_utils.monitoring.set_custom_attribute')
    def test_request_auth_type_guess_anonymous_attribute(self, mock_set_custom_attribute):
        self.request.user = AnonymousUser()

        self.middleware.process_response(self.request, None)
        mock_set_custom_attribute.assert_called_once_with('request_auth_type_guess', 'unauthenticated')



"""
API for CSRF application.
"""

from django.middleware.csrf import get_token
from rest_framework.permissions import AllowAny
from rest_framework.response import Response
from rest_framework.views import APIView


class CsrfTokenView(APIView):
    """
        **Use Case**

            Allows frontend apps to obtain a CSRF token from the Django
            service in order to make POST, PUT, and DELETE requests to
            API endpoints hosted on the service.

        **Behavior**

            GET /csrf/api/v1/token
            >>> {
            >>>     "csrfToken": "abcdefg1234567"
            >>> }
    """
    # AllowAny keeps possible default of DjangoModelPermissions from being used.
    permission_classes = (AllowAny,)

    def get(self, request):
        """
        GET /csrf/api/v1/token
        """
        return Response({'csrfToken': get_token(request)})



""" Tests for Bearer authentication class. """
import json
from unittest import mock

import httpretty
from django.contrib.auth import get_user_model
from django.test import RequestFactory, TestCase, override_settings
from requests import RequestException
from rest_framework.exceptions import AuthenticationFailed

from edx_rest_framework_extensions.auth.bearer.authentication import (
    BearerAuthentication,
)
from edx_rest_framework_extensions.tests import factories


OAUTH2_USER_INFO_URL = 'http://example.com/oauth2/user_info/'
USER_INFO = {
    'username': 'jdoe',
    'first_name': 'Jane',
    'last_name': 'Doê',
    'email': 'jdoe@example.com',
}
User = get_user_model()


class AccessTokenMixin:
    """ Test mixin for dealing with OAuth2 access tokens. """
    DEFAULT_TOKEN = 'abc123'

    def mock_user_info_response(self, status=200, username=None):
        """ Mock the user info endpoint response of the OAuth2 provider. """

        username = username or USER_INFO['username']
        data = {
            'family_name': USER_INFO['last_name'],
            'preferred_username': username,
            'given_name': USER_INFO['first_name'],
            'email': USER_INFO['email'],
        }

        httpretty.register_uri(
            httpretty.GET,
            OAUTH2_USER_INFO_URL,
            body=json.dumps(data),
            content_type='application/json',
            status=status
        )


@override_settings(EDX_DRF_EXTENSIONS={'OAUTH2_USER_INFO_URL': OAUTH2_USER_INFO_URL})
class BearerAuthenticationTests(AccessTokenMixin, TestCase):
    """ Tests for the BearerAuthentication class. """
    TOKEN_NAME = 'Bearer'

    def setUp(self):
        super().setUp()
        self.auth = BearerAuthentication()
        self.factory = RequestFactory()

    def create_authenticated_request(self, token=AccessTokenMixin.DEFAULT_TOKEN, token_name=TOKEN_NAME):
        """ Returns a Request with the authorization set using the specified values. """
        auth_header = f'{token_name} {token}'
        request = self.factory.get('/', HTTP_AUTHORIZATION=auth_header)
        return request

    def assert_user_authenticated(self):
        """ Assert a user can be authenticated with a bearer token. """
        user = factories.UserFactory()
        self.mock_user_info_response(username=user.username)

        request = self.create_authenticated_request()
        self.assertEqual(self.auth.authenticate(request), (user, self.DEFAULT_TOKEN))

    def assert_authentication_failed(self, token=AccessTokenMixin.DEFAULT_TOKEN, token_name=TOKEN_NAME):
        """ Assert authentication fails for a generated request. """
        request = self.create_authenticated_request(token=token, token_name=token_name)
        self.assertRaises(AuthenticationFailed, self.auth.authenticate, request)

    def test_authenticate_header(self):
        """ The method should return the string Bearer. """
        self.assertEqual(self.auth.authenticate_header(self.create_authenticated_request()), 'Bearer')

    @override_settings(EDX_DRF_EXTENSIONS={'OAUTH2_USER_INFO_URL': None})
    def test_authenticate_no_user_info_url(self):
        """ If the setting OAUTH2_USER_INFO_URL is not set, the method returns None. """

        # Empty value
        self.assertIsNone(self.auth.authenticate(self.create_authenticated_request()))

        # Missing value
        with override_settings(EDX_DRF_EXTENSIONS={}):
            self.assertIsNone(self.auth.authenticate(self.create_authenticated_request()))

    def test_authenticate_invalid_token(self):
        """ If no token is supplied, or if the token contains spaces, the method should raise an exception. """

        # Missing token
        self.assert_authentication_failed(token='')

        # Token with spaces
        self.assert_authentication_failed(token='abc 123 456')

    def test_authenticate_invalid_token_name(self):
        """ If the token name is not Bearer, the method should return None. """
        request = self.create_authenticated_request(token_name='foobar')
        self.assertIsNone(self.auth.authenticate(request))

    @httpretty.activate
    def test_authenticate_inactive_user(self):
        """ If the user matching the access token is inactive, the method should raise an exception. """
        user = factories.UserFactory(is_active=False)
        self.mock_user_info_response(username=user.username)
        self.assert_authentication_failed()

    @httpretty.activate
    def test_authenticate_invalid_token_response(self):
        """ If the user info endpoint does not return HTTP 200, the method should return raise an exception. """
        self.mock_user_info_response(status=400)
        self.assert_authentication_failed()

    @httpretty.activate
    def test_authenticate(self):
        """ If the access token is valid, the user exists, and is active, a tuple containing
        the user and token should be returned.
        """
        self.assert_user_authenticated()

    @httpretty.activate
    def test_authenticate_as_new_user(self):
        """ Verify a new user is created. """
        self.mock_user_info_response()
        request = self.create_authenticated_request()
        actual_user, actual_token = self.auth.authenticate(request)

        self.assertEqual(actual_token, self.DEFAULT_TOKEN)
        self.assertEqual(actual_user, User.objects.get(username=USER_INFO['username']))

    @httpretty.activate
    def test_authenticate_user_creation_with_existing_user(self):
        """ Verify an existing user is returned, if the user already exists. """
        user = factories.UserFactory(username=USER_INFO['username'])
        self.mock_user_info_response()
        request = self.create_authenticated_request()
        actual_user, actual_token = self.auth.authenticate(request)

        self.assertEqual(actual_token, self.DEFAULT_TOKEN)
        self.assertEqual(actual_user, user)

    @httpretty.activate
    def test_authenticate_user_creation_with_request_status_failure(self):
        """ Verify authentication fails if the request to retrieve user info returns a non-200 status. """
        original_user_count = User.objects.all().count()
        self.mock_user_info_response(status=401)
        request = self.create_authenticated_request()

        self.assertRaises(AuthenticationFailed, self.auth.authenticate, request)
        self.assertEqual(User.objects.all().count(), original_user_count)

    def test_authenticate_user_creation_with_request_exception(self):
        """ Verify authentication fails if the request to retrieve user info raises an exception. """
        original_user_count = User.objects.all().count()
        request = self.create_authenticated_request()

        with mock.patch('requests.get', mock.Mock(side_effect=RequestException)):
            self.assertRaises(AuthenticationFailed, self.auth.authenticate, request)

        self.assertEqual(User.objects.all().count(), original_user_count)



"""
Unit tests for jwt cookies module.
"""
from unittest import mock

import ddt
from django.test import TestCase, override_settings

from edx_rest_framework_extensions.auth.jwt.decoder import jwt_decode_handler
from edx_rest_framework_extensions.auth.jwt.tests.utils import (
    generate_jwt_token,
    generate_latest_version_payload,
)
from edx_rest_framework_extensions.tests.factories import UserFactory

from .. import cookies


@ddt.ddt
class TestJwtAuthCookies(TestCase):
    @ddt.data(
        (cookies.jwt_cookie_name, 'JWT_AUTH_COOKIE', 'custom-jwt-cookie-name'),
        (cookies.jwt_cookie_header_payload_name, 'JWT_AUTH_COOKIE_HEADER_PAYLOAD', 'custom-jwt-header-payload-name'),
        (cookies.jwt_cookie_signature_name, 'JWT_AUTH_COOKIE_SIGNATURE', 'custom-jwt-signature-name'),
    )
    @ddt.unpack
    def test_get_setting_value(self, jwt_cookie_func, setting_name, setting_value):
        with override_settings(JWT_AUTH={setting_name: setting_value}):
            self.assertEqual(jwt_cookie_func(), setting_value)

    @ddt.data(
        (cookies.jwt_cookie_name, 'edx-jwt-cookie'),
        (cookies.jwt_cookie_header_payload_name, 'edx-jwt-cookie-header-payload'),
        (cookies.jwt_cookie_signature_name, 'edx-jwt-cookie-signature'),
    )
    @ddt.unpack
    def test_get_default_value(self, jwt_cookie_func, expected_default_value):
        self.assertEqual(jwt_cookie_func(), expected_default_value)

    def test_get_decoded_jwt_from_existing_cookie(self):
        user = UserFactory()
        payload = generate_latest_version_payload(user)
        jwt = generate_jwt_token(payload)
        expected_decoded_jwt = jwt_decode_handler(jwt)

        mock_request_with_cookie = mock.Mock(COOKIES={'edx-jwt-cookie': jwt})

        decoded_jwt = cookies.get_decoded_jwt(mock_request_with_cookie)
        self.assertEqual(expected_decoded_jwt, decoded_jwt)

    def test_get_decoded_jwt_when_no_cookie(self):
        mock_request = mock.Mock(COOKIES={})

        self.assertIsNone(cookies.get_decoded_jwt(mock_request))



"""
App for creating and distributing CSRF tokens to frontend applications.
"""

from django.apps import AppConfig


class CsrfAppConfig(AppConfig):
    """Configuration for the csrf application."""

    name = 'csrf'
    verbose_name = 'CSRF'



"""
JWT Authentication cookie utilities.
"""

from django.conf import settings

from edx_rest_framework_extensions.auth.jwt.decoder import configured_jwt_decode_handler


def jwt_cookie_name():
    return settings.JWT_AUTH.get('JWT_AUTH_COOKIE') or 'edx-jwt-cookie'


def jwt_cookie_header_payload_name():
    return settings.JWT_AUTH.get('JWT_AUTH_COOKIE_HEADER_PAYLOAD') or 'edx-jwt-cookie-header-payload'


def jwt_cookie_signature_name():
    return settings.JWT_AUTH.get('JWT_AUTH_COOKIE_SIGNATURE') or 'edx-jwt-cookie-signature'


def get_decoded_jwt(request):
    """
    Grab jwt from jwt cookie in request if possible.

    Returns a decoded jwt dict if it can be found.
    Returns None if the jwt is not found.
    """
    jwt_cookie = request.COOKIES.get(jwt_cookie_name(), None)

    if not jwt_cookie:
        return None
    return configured_jwt_decode_handler(jwt_cookie)



#!/usr/bin/env python
from setuptools import setup, find_packages

import edx_rest_framework_extensions


def is_requirement(line):
    """
    Return True if the requirement line is a package requirement;
    that is, it is not blank, a comment, or editable.
    """
    # Remove whitespace at the start/end of the line
    line = line.strip()

    # Skip blank lines, comments, and editable installs
    return not (
        line == '' or
        line.startswith('-r') or
        line.startswith('#') or
        line.startswith('-e') or
        line.startswith('git+') or
        line.startswith('-c')
    )


def load_requirements(*requirements_paths):
    """
    Load all requirements from the specified requirements files.
    Returns a list of requirement strings.
    """
    requirements = set()
    for path in requirements_paths:
        requirements.update(
            line.strip() for line in open(path).readlines()
            if is_requirement(line)
        )
    return list(requirements)


setup(
    name='edx-drf-extensions',
    version=edx_rest_framework_extensions.__version__,
    description='edX extensions of Django REST Framework',
    author='edX',
    author_email='oscm@edx.org',
    url='https://github.com/edx/edx-drf-extensions',
    license='Apache 2.0',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Framework :: Django',
        'Framework :: Django :: 2.2',
        'Framework :: Django :: 3.0',
        'Framework :: Django :: 3.1',
        'Framework :: Django :: 3.2',
    ],
    packages=find_packages(exclude=["tests"]),
    install_requires=load_requirements('requirements/base.in'),
    tests_require=load_requirements('requirements/test.in'),
)
